import sys
import os
import tempfile
from . import command, getTestDataPath, MockSignal
from pytestqt import qtbot


def test_commandline_classic_export(qtbot, command):
    """Run Qt event loop and create a video in the system /tmp or /temp"""
    soundFile = getTestDataPath("inputfiles/test.ogg")
    outputDir = tempfile.mkdtemp(prefix="avp-export-")
    outputFilename = os.path.join(outputDir, "output.mp4")
    sys.argv = [
        "",
        "-c",
        "0",
        "classic",
        "color=255,255,255",
        "-i",
        soundFile,
        "-o",
        outputFilename,
    ]
    command.parseArgs()
    # Command object now has a video_thread Worker which is exporting the video

    with qtbot.waitSignal(command.worker.videoCreated, timeout=10000):
        """
        Wait until videoCreated is emitted by the video_thread Worker
        or until 10 second timeout has passed
        """
        print(f"Test Video created at {outputFilename}")

    assert os.path.exists(outputFilename)
    # output video should be at least 200kb
    assert os.path.getsize(outputFilename) > 200000
